# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Entities to be used with an overlay on Operation or DataType, are defined here.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import numpy
import six
from tvb.basic.config.utils import EnhancedDictionary
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.utils import date2string


class CommonDetails(EnhancedDictionary):
    """
    Enhanced dictionary holding details about an entity.
    It also contains metadata on how this details will be displayed in UI (disabled/readonly/hidden input field).
    """

    CODE_GID = "gid"
    CODE_OPERATION_TAG = "operation_label"
    CODE_OPERATION_GROUP_ID = 'operation_group_id'
    CODE_OPERATION_GROUP_NAME = 'operation_group_name'

    def __init__(self):

        super(CommonDetails, self).__init__()
        self.metadata = dict()
        self.scientific_details = dict()

        self.gid = None
        self.metadata[self.CODE_GID] = {"name": "Unique Global Identifier", "readonly": "True"}

        self.author = None
        self.metadata['author'] = {"name": "Author", "disabled": "True"}

        self.operation_type = None
        self.metadata['operation_type'] = {"name": "Operation Algorithm", "disabled": "True"}

        self.operation_label = None
        self.metadata[self.CODE_OPERATION_TAG] = {"name": "Operation Label"}

        self.count = 1
        self.metadata['count'] = {"name": "No. of entities in group", "disabled": "True"}

        self.operation_group_id = None
        self.metadata[self.CODE_OPERATION_GROUP_ID] = {"name": "groupId", "hidden": "True"}

        self.operation_group_name = None
        self.metadata[self.CODE_OPERATION_GROUP_NAME] = {"name": "operationGroupName", "hidden": "True"}

        self.burst_name = None
        self.metadata['burst_name'] = {"name": "Simulation", "disabled": "True"}


    def add_scientific_fields(self, extra_fields_dict):
        """
        :param extra_fields_dict a dictionary of fields to be added in a distinct category.
        """
        for key, value in six.iteritems(extra_fields_dict):
            if value is not None and isinstance(value, numpy.ndarray):
                # Make sure arrays are displayed in a compatible form: [1, 2, 3]
                value = str(value.tolist())
            elif value is not None and isinstance(value, list) and len(value) > 0 and isinstance(value[0], str):
                value = [str(v) for v in value]
            self.scientific_details[key] = {"name": key, "value": str(value), "disabled": "True"}


    @property
    def meta_attributes_list(self):
        """
        A list with all attributes for a DataType which are expected to be submitted from UI.
        """
        result = list(self)
        result.remove('metadata')
        result.remove('scientific_details')
        return result


    @staticmethod
    def compute_operation_name(category_name, algorithm_name):
        """
        Compute Operation display name, based on Operation Category and Algorithm.
        """
        display_name = ''
        if category_name:
            display_name = category_name
        if algorithm_name:
            display_name = display_name + "->" + algorithm_name
        return display_name


    def get_ui_fields(self):
        """
        :returns: list of dictionaries to be used for UI display. \
                  Each entry in the list will be displayed as a UI group of fields.
        """
        framework_metadata = self.metadata
        for key in self.meta_attributes_list:
            framework_metadata[key]['value'] = getattr(self, key)
        if len(self.scientific_details) > 0:
            return [self.scientific_details, framework_metadata]
        return [framework_metadata]


class OperationOverlayDetails(CommonDetails):
    """
    Entity used for displaying in an overlay details about an operation.
    """


    def __init__(self, operation, user_display_name, count_inputs, count_results, burst, no_of_op_in_group, op_pid):
        super(OperationOverlayDetails, self).__init__()

        self.operation_id = operation.id
        self.metadata['operation_id'] = {"name": "Entity id", "disabled": "True"}

        self.operation_status = operation.status
        self.metadata['operation_status'] = {"name": "Status", "disabled": "True"}

        self.start_date = date2string(operation.start_date)
        self.metadata['start_date'] = {"name": "Start date", "disabled": "True"}

        self.end_date = "None" if operation.completion_date is None else date2string(operation.completion_date)
        self.metadata['end_date'] = {"name": "End date", "disabled": "True"}

        self.result_datatypes = count_results
        self.metadata['result_datatypes'] = {"name": "No. of resulted DataTypes", "disabled": "True"}

        self.input_datatypes = count_inputs
        self.metadata['input_datatypes'] = {"name": "No. of input DataTypes", "disabled": "True"}

        # If the operation was executed in another process or as a cluster then show the pid/job_id
        if op_pid is not None:
            if op_pid.pid is not None:
                self.pid = op_pid.pid
                self.metadata['pid'] = {"name": "Process pid", "disabled": "True"}
            elif op_pid.job_id is not None:
                self.job_id = op_pid.job_id
                self.metadata['job_id'] = {"name": "Cluster job Id", "disabled": "True"}

        # Now set/update generic fields
        self.gid = operation.gid
        self.author = user_display_name
        self.count = no_of_op_in_group
        self.burst_name = burst.name if burst is not None else ""
        self.operation_type = self.compute_operation_name(operation.algorithm.algorithm_category.displayname,
                                                          operation.algorithm.displayname)

        group_id = operation.operation_group.id if operation.operation_group is not None else None
        self.operation_group_id = group_id

        self.operation_label = (operation.operation_group.name if operation.operation_group is not None
                                else operation.user_group)
        self.metadata[self.CODE_OPERATION_TAG]["disabled"] = 'True'

        group_value = operation.operation_group.name if operation.operation_group is not None else None
        self.operation_group_name = group_value
        self.metadata[self.CODE_OPERATION_GROUP_NAME]["disabled"] = 'True'


class DataTypeOverlayDetails(CommonDetails):
    """
    Entity used for displaying in an overlay details about a DataType.
    """

    DATA_SUBJECT = "subject"
    DATA_STATE = "data_state"
    DATA_TITLE = "datatype_title"
    DATA_TAG_1 = "datatype_tag_1"
    DATA_TAG_2 = "datatype_tag_2"
    DATA_TAG_3 = "datatype_tag_3"
    DATA_TAG_4 = "datatype_tag_4"
    DATA_TAG_5 = "datatype_tag_5"


    def __init__(self):

        super(DataTypeOverlayDetails, self).__init__()

        self.data_state = None
        self.metadata[self.DATA_STATE] = {"name": "State"}

        self.datatype_title = None
        self.metadata[self.DATA_TITLE] = {"name": "Display Name", "disabled": "True"}

        self.subject = None
        self.metadata[self.DATA_SUBJECT] = {"name": "Subject"}

        self.data_type = None
        self.metadata['data_type'] = {"name": "Data Type", "disabled": "True"}

        self.data_type_id = None
        self.metadata['data_type_id'] = {"name": "Entity id", "disabled": "True"}

        self.create_date = None
        self.metadata['create_date'] = {"name": "Creation Date", "disabled": "True"}

        self.datatype_tag_1 = None
        self.metadata[self.DATA_TAG_1] = {"name": "DataType Tag 1"}

        self.datatype_tag_2 = None
        self.metadata[self.DATA_TAG_2] = {"name": "DataType Tag 2"}

        self.datatype_tag_3 = None
        self.metadata[self.DATA_TAG_3] = {"name": "DataType Tag 3"}

        self.datatype_tag_4 = None
        self.metadata[self.DATA_TAG_4] = {"name": "DataType Tag 4"}

        self.datatype_tag_5 = None
        self.metadata[self.DATA_TAG_5] = {"name": "DataType Tag 5"}

        self.datatype_size = None
        self.metadata['datatype_size'] = {"name": "Size on Disk (KB)", "disabled": "True"}


    def fill_from_datatype(self, datatype_result, parent_burst):
        """
        Fill current dictionary with information from a loaded DB DataType.
        :param datatype_result DB loaded DataType
        :param parent_burst Burst entity in which current dataType was generated
        """

        self.gid = datatype_result.gid
        self.data_type_id = datatype_result.id
        self.data_state = datatype_result.state
        self.datatype_title = datatype_result.display_name
        self.subject = datatype_result.subject
        self.data_type = datatype_result.display_type
        self.datatype_tag_1 = datatype_result.user_tag_1
        self.datatype_tag_2 = datatype_result.user_tag_2
        self.datatype_tag_3 = datatype_result.user_tag_3
        self.datatype_tag_4 = datatype_result.user_tag_4
        self.datatype_tag_5 = datatype_result.user_tag_5
        self.datatype_size = datatype_result.disk_size
        self.author = datatype_result.parent_operation.user.display_name

        parent_algorithm = datatype_result.parent_operation.algorithm
        operation_name = self.compute_operation_name(parent_algorithm.algorithm_category.displayname,
                                                     parent_algorithm.displayname)
        self.operation_type = operation_name

        create_date_str = ''
        if datatype_result.parent_operation.completion_date is not None:
            create_date_str = date2string(datatype_result.parent_operation.completion_date)
        self.create_date = create_date_str

        if parent_burst is not None:
            self.burst_name = parent_burst.name
        else:
            self.burst_name = ''

        # Populate Group attributes
        if isinstance(datatype_result, DataTypeGroup):
            self.count = datatype_result.count_results
            self.operation_group_name = datatype_result.parent_operation.operation_group.name
            self.operation_label = datatype_result.parent_operation.operation_group.name
            self.operation_group_id = datatype_result.parent_operation.operation_group.id
        else:
            self.operation_label = datatype_result.parent_operation.user_group

        # Populate Scientific attributes
        self.add_scientific_fields(datatype_result.summary_info)


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Robert Vincze <robert.vincze@codemart.ro>
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import os
import numpy
from abc import ABCMeta
import pyAesCrypt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from scipy import io as scipy_io
from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.core.adapters.abcadapter import AdapterLaunchModeEnum, ABCAdapterForm, ABCAdapter
from tvb.core.adapters.exceptions import LaunchException
from tvb.core.neotraits.forms import StrField, TraitUploadField
from tvb.core.neotraits.uploader_view_model import UploaderViewModel


ENCRYPTED_PASSWORD_NAME = 'encrypted_password.pem'
ENCRYPTED_DATA_SUFFIX = '_encrypted'
DECRYPTED_DATA_SUFFIX = '_decrypted'


class ABCUploaderForm(ABCAdapterForm):

    def __init__(self):
        super(ABCUploaderForm, self).__init__()
        self.subject_field = StrField(UploaderViewModel.data_subject, name='Data_Subject')
        # Show Encryption field only when the current TVB installation is capable of decryption
        supports_encrypted_files = (TvbProfile.current.UPLOAD_KEY_PATH is not None
                                    and os.path.exists(TvbProfile.current.UPLOAD_KEY_PATH))
        if supports_encrypted_files:
            self.encrypted_aes_key = TraitUploadField(UploaderViewModel.encrypted_aes_key, '.pem',
                                                      'encrypted_aes_key')

    @staticmethod
    def get_required_datatype():
        return None

    @staticmethod
    def get_filters():
        return None

    @staticmethod
    def get_input_name():
        return None

    def get_upload_field_names(self):
        for field in self.trait_fields:
            if isinstance(field, TraitUploadField):
                yield field.trait_attribute.field_name


class ABCUploader(ABCAdapter, metaclass=ABCMeta):
    """
    Base class of the uploading algorithms
    """
    LOGGER = get_logger(__name__)
    launch_mode = AdapterLaunchModeEnum.SYNC_DIFF_MEM

    def _prelaunch(self, operation, view_model, available_disk_space=0):
        """
        Before going with the usual prelaunch, get from input parameters the 'subject'.
        """
        self.generic_attributes.subject = view_model.data_subject

        trait_upload_field_names = list(self.get_form_class().get_upload_information().keys())
        if view_model.encrypted_aes_key is not None:
            for upload_field_name in trait_upload_field_names:
                self._decrypt_content(view_model, upload_field_name)

        return ABCAdapter._prelaunch(self, operation, view_model, available_disk_space)

    @staticmethod
    def get_path_to_encrypt(input_path):
        start_extension = input_path.rfind('.')
        path_to_encrypt = input_path[:start_extension]
        extension = input_path[start_extension:]

        return path_to_encrypt + ENCRYPTED_DATA_SUFFIX + extension

    @staticmethod
    def encrypt_password(public_key, symmetric_key):

        encrypted_symmetric_key = public_key.encrypt(
            symmetric_key,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        return encrypted_symmetric_key

    @staticmethod
    def save_encrypted_password(encrypted_password, path_to_encrypted_password):

        with open(os.path.join(path_to_encrypted_password, ENCRYPTED_PASSWORD_NAME), 'wb') as f:
            f.write(encrypted_password)

    @staticmethod
    def _decrypt_content(view_model, trait_upload_field_name):
        if TvbProfile.current.UPLOAD_KEY_PATH is None or not os.path.exists(TvbProfile.current.UPLOAD_KEY_PATH):
            raise LaunchException("We can not process Encrypted files at this moment, "
                                  "due to missing PK for decryption! Please contact the administrator!")

        upload_path = getattr(view_model, trait_upload_field_name)

        # Get the encrypted password
        with open(view_model.encrypted_aes_key, 'rb') as f:
            encrypted_password = f.read()

        # Read the private key
        with open(TvbProfile.current.UPLOAD_KEY_PATH, "rb") as key_file:
            private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )

        # Decrypt the password using the private key
        decrypted_password = private_key.decrypt(
            encrypted_password,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )

        decrypted_password = decrypted_password.decode()

        # Get path to decrypted file
        decrypted_download_path = upload_path.replace(ENCRYPTED_DATA_SUFFIX, DECRYPTED_DATA_SUFFIX)

        # Use the decrypted password to decrypt the message
        pyAesCrypt.decryptFile(upload_path, decrypted_download_path, decrypted_password,
                               TvbProfile.current.hpc.CRYPT_BUFFER_SIZE)
        view_model.__setattr__(trait_upload_field_name, decrypted_download_path)

    def get_required_memory_size(self, view_model):
        """
        Return the required memory to run this algorithm.
        As it is an upload algorithm and we do not have information about data, we can not approximate this.
        """
        return -1

    def get_required_disk_size(self, view_model):
        """
        As it is an upload algorithm and we do not have information about data, we can not approximate this.
        """
        return 0

    @staticmethod
    def read_list_data(full_path, dimensions=None, dtype=numpy.float64, skiprows=0, usecols=None):
        """
        Read numpy.array from a text file or a npy/npz file.
        """
        try:
            if full_path.endswith(".npy") or full_path.endswith(".npz"):
                array_result = numpy.load(full_path)
            else:
                array_result = numpy.loadtxt(full_path, dtype=dtype, skiprows=skiprows, usecols=usecols)
            if dimensions:
                return array_result.reshape(dimensions)
            return array_result
        except ValueError as exc:
            file_ending = os.path.split(full_path)[1]
            exc.args = (exc.args[0] + " In file: " + file_ending,)
            raise

    @staticmethod
    def read_matlab_data(path, matlab_data_name=None):
        """
        Read array from matlab file.
        """
        try:
            matlab_data = scipy_io.matlab.loadmat(path)
        except NotImplementedError:
            ABCUploader.LOGGER.error("Could not read Matlab content from: " + path)
            ABCUploader.LOGGER.error("Matlab files must be saved in a format <= -V7...")
            raise

        try:
            return matlab_data[matlab_data_name]
        except KeyError:
            def double__(n):
                n = str(n)
                return n.startswith('__') and n.endswith('__')

            available = [s for s in matlab_data if not double__(s)]
            raise KeyError("Could not find dataset named %s. Available datasets: %s" % (matlab_data_name, available))

    @staticmethod
    def get_upload_information():
        return NotImplementedError



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Paula Popa <paula.popa@codemart.ro>
"""

import os
import uuid
import typing
from tvb.adapters.analyzers.metrics_group_timeseries import TimeseriesMetricsAdapterModel, \
    TimeseriesMetricsAdapter
from tvb.adapters.datatypes.db.mapped_value import DatatypeMeasureIndex
from tvb.adapters.datatypes.db.simulation_history import SimulationHistoryIndex
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.adapters.simulator.simulator_adapter import SimulatorAdapter, SimulatorAdapterModel
from tvb.basic.neotraits.api import HasTraits
from tvb.config import ALGORITHMS
from tvb.core.neocom import h5
from tvb.core.neotraits.h5 import H5File, ViewModelH5
from tvb.core.services.backend_clients.hpc_scheduler_client import HPCSchedulerClient


class HPCSimulatorAdapter(SimulatorAdapter):

    def __init__(self, storage_path, is_group_launch):
        super(HPCSimulatorAdapter, self).__init__()
        self.storage_path = storage_path
        self.is_group_launch = is_group_launch

    def _prelaunch(self, operation, view_model, available_disk_space=0):
        self.available_disk_space = available_disk_space
        super(HPCSimulatorAdapter, self)._prelaunch(operation, view_model, available_disk_space)

    def get_output(self):
        return [TimeSeriesIndex, SimulationHistoryIndex, DatatypeMeasureIndex]

    def load_traited_by_gid(self, data_gid):
        # type: (uuid.UUID) -> HasTraits
        """
        Load a generic HasTraits instance, specified by GID.
        """
        trait, _ = h5.load_with_links_from_dir(self.storage_path, data_gid)
        return trait

    def load_with_references(self, dt_gid):
        # type: (typing.Union[uuid.UUID, str]) -> HasTraits
        dt, _ = h5.load_with_references_from_dir(self.storage_path, dt_gid)
        return dt

    def _try_load_region_mapping(self):
        return None, None

    def _is_group_launch(self):
        """
        Return true if this adapter is launched from a group of operations
        """
        return self.is_group_launch

    def _get_output_path(self):
        output_path = os.path.join(self.storage_path, HPCSchedulerClient.OUTPUT_FOLDER)
        if not os.path.isdir(output_path):
            os.mkdir(output_path)
        return output_path

    def extract_operation_data(self, operation=None):
        """
        Do nothing for HPC run.
        :param operation: None
        """

    def _update_operation_entity(self, operation, required_disk_space):
        """
        """

    def _capture_operation_results(self, result):
        """
        Update h5 files with generic attributes
        """
        for file in os.listdir(self._get_output_path()):
            path = os.path.join(self._get_output_path(), file)
            if issubclass(H5File.h5_class_from_file(path), ViewModelH5):
                continue
            with H5File.from_file(path) as f:
                f.store_generic_attributes(self.generic_attributes)
        return "", 2

    def _ensure_enough_resources(self, available_disk_space, view_model):
        return 0

    def launch(self, view_model):
        # type: (SimulatorAdapterModel) -> [TimeSeriesIndex, SimulationHistoryIndex]
        simulation_results = super(HPCSimulatorAdapter, self).launch(view_model)

        if not self.is_group_launch:
            return simulation_results

        for dt in simulation_results:
            if issubclass(type(dt), TimeSeriesIndex):
                self._compute_metrics_for_pse_launch(dt)

        return simulation_results

    def _compute_metrics_for_pse_launch(self, time_series_index):
        # type: (TimeSeriesIndex) -> [DatatypeMeasureIndex]
        metric_vm = TimeseriesMetricsAdapterModel()
        metric_vm.time_series = time_series_index.gid
        metric_vm.algorithms = tuple(ALGORITHMS.keys())
        h5.store_view_model(metric_vm, self._get_output_path())
        metric_adapter = HPCTimeseriesMetricsAdapter(self._get_output_path(), time_series_index)
        metric_adapter._prelaunch(None, metric_vm, self.available_disk_space)


class HPCTimeseriesMetricsAdapter(TimeseriesMetricsAdapter):

    def __init__(self, storage_path, input_time_series_index):
        super(HPCTimeseriesMetricsAdapter, self).__init__()
        self.storage_path = storage_path
        self.input_time_series_index = input_time_series_index

    def configure(self, view_model):
        # type: (TimeseriesMetricsAdapterModel) -> None
        """
        Store the input shape to be later used to estimate memory usage.
        """
        self.input_shape = (self.input_time_series_index.data_length_1d,
                            self.input_time_series_index.data_length_2d,
                            self.input_time_series_index.data_length_3d,
                            self.input_time_series_index.data_length_4d)

    def load_traited_by_gid(self, data_gid):
        # type: (typing.Union[uuid.UUID, str]) -> HasTraits
        """
        Load a generic HasTraits instance, specified by GID.
        """
        trait, _ = h5.load_with_links_from_dir(self.storage_path, data_gid)
        return trait

    def _get_output_path(self):
        return self.storage_path

    def extract_operation_data(self, operation=None):
        """
        Do nothing for HPC run.
        :param operation: None
        """

    def _update_operation_entity(self, operation, required_disk_space):
        """
        """

    def _capture_operation_results(self, result):
        """
        """
        return "", 1

    def _ensure_enough_resources(self, available_disk_space, view_model):
        return 0



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.adapters.datatypes.h5.spectral_h5 import DataTypeMatrixH5
from tvb.adapters.datatypes.h5.structural_h5 import VolumetricDataMixin
from tvb.core.neotraits.h5 import H5File, DataSet, Reference
from tvb.datatypes.region_mapping import RegionMapping, RegionVolumeMapping


class RegionMappingH5(H5File):

    def __init__(self, path):
        super(RegionMappingH5, self).__init__(path)
        self.array_data = DataSet(RegionMapping.array_data, self)
        self.connectivity = Reference(RegionMapping.connectivity, self)
        self.surface = Reference(RegionMapping.surface, self)

    def get_region_mapping_slice(self, start_idx, end_idx):
        """
        Get a slice of the region mapping as used by the region viewers.
        For each vertex on the surface, alpha-indices will be the closest
        region-index
        :param start_idx: vertex index on the surface
        :param end_idx: vertex index on the surface
        :return: NumPy array with [closest_reg_idx ...]
        """
        return self.array_data.load()[int(start_idx): int(end_idx)].T


class RegionVolumeMappingH5(VolumetricDataMixin, DataTypeMatrixH5):

    def __init__(self, path):
        super(RegionVolumeMappingH5, self).__init__(path)
        self.array_data = DataSet(RegionVolumeMapping.array_data, self)
        self.connectivity = Reference(RegionVolumeMapping.connectivity, self)
        self.volume = Reference(RegionVolumeMapping.volume, self)

    def read_data_shape(self):
        """
        The shape of the data
        """
        return self.array_data.shape

    def read_data_slice(self, data_slice):
        """
        Expose chunked-data access.
        """
        return self.array_data[data_slice]



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from sqlalchemy import Column, Integer, ForeignKey, String
from sqlalchemy.orm import relationship
from tvb.datatypes.projections import ProjectionMatrix
from tvb.adapters.datatypes.db.sensors import SensorsIndex
from tvb.adapters.datatypes.db.surface import SurfaceIndex
from tvb.core.entities.model.model_datatype import DataType


class ProjectionMatrixIndex(DataType):
    id = Column(Integer, ForeignKey(DataType.id), primary_key=True)

    projection_type = Column(String, nullable=False)

    fk_brain_skull_gid = Column(String(32), ForeignKey(SurfaceIndex.gid),
                                nullable=not ProjectionMatrix.brain_skull.required)
    brain_skull = relationship(SurfaceIndex, foreign_keys=fk_brain_skull_gid,
                               primaryjoin=SurfaceIndex.gid == fk_brain_skull_gid, cascade='none')

    fk_skull_skin_gid = Column(String(32), ForeignKey(SurfaceIndex.gid), nullable=not ProjectionMatrix.skull_skin.required)
    skull_skin = relationship(SurfaceIndex, foreign_keys=fk_skull_skin_gid, primaryjoin=SurfaceIndex.gid == fk_skull_skin_gid,
                              cascade='none')

    fk_skin_air_gid = Column(String(32), ForeignKey(SurfaceIndex.gid), nullable=not ProjectionMatrix.skin_air.required)
    skin_air = relationship(SurfaceIndex, foreign_keys=fk_skin_air_gid, primaryjoin=SurfaceIndex.gid == fk_skin_air_gid,
                            cascade='none')

    fk_source_gid = Column(String(32), ForeignKey(SurfaceIndex.gid), nullable=not ProjectionMatrix.sources.required)
    source = relationship(SurfaceIndex, foreign_keys=fk_source_gid, primaryjoin=SurfaceIndex.gid == fk_source_gid,
                          cascade='none')

    fk_sensors_gid = Column(String(32), ForeignKey(SensorsIndex.gid), nullable=not ProjectionMatrix.sensors.required)
    sensors = relationship(SensorsIndex, foreign_keys=fk_sensors_gid, primaryjoin=SensorsIndex.gid == fk_sensors_gid,
                           cascade='none')

    def fill_from_has_traits(self, datatype):
        # type: (ProjectionMatrix)  -> None
        super(ProjectionMatrixIndex, self).fill_from_has_traits(datatype)
        self.projection_type = datatype.projection_type
        if datatype.brain_skull is not None:
            self.fk_brain_skull_gid = datatype.brain_skull.gid.hex
        if datatype.skull_skin is not None:
            self.fk_skull_skin_gid = datatype.skull_skin.gid.hex
        if datatype.skin_air is not None:
            self.fk_skin_air_gid = datatype.skin_air.gid.hex
        self.fk_sensors_gid = datatype.sensors.gid.hex
        self.fk_source_gid = datatype.sources.gid.hex

    def get_subtype_attr(self):
        return self.projection_type

    @property
    def display_name(self):
        """
        Overwrite from superclass and add subtype
        """
        previous = "ProjectionMatrix"
        return previous + " - " + str(self.projection_type)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Populate Surface fields after 1.2.2, in version 1.2.3.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from tvb.basic.logger.builder import get_logger
from tvb.core.entities.storage import dao
from tvb.core.entities import model
from tvb.datatypes.surfaces import Surface


LOGGER = get_logger(__name__)

def update():
    """
    Update Surface metadata
    """

    try:
        all_surfaces = dao.get_generic_entity(model.DataType, "tvb.datatypes.surfaces", "module")
        for srf in all_surfaces:
            surface = dao.get_datatype_by_gid(srf.gid)
            if isinstance(surface, Surface):
                surface.configure()
                dao.store_entity(surface)
                surface.persist_full_metadata()
    except Exception:
        LOGGER.exception("Could update Surface entities!")



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

from tvb.adapters.datatypes.db.graph import ConnectivityMeasureIndex
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import Attr
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.core.adapters.exceptions import ParseException, LaunchException
from tvb.core.entities.storage import transactional
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import TraitUploadField, StrField, TraitDataTypeSelectField
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str, DataTypeGidAttr
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.graph import ConnectivityMeasure


class ConnectivityMeasureImporterModel(UploaderViewModel):
    data_file = Str(
        label='Connectivity measure file (.mat format)'
    )

    dataset_name = Attr(
        field_type=str,
        default='M',
        label='Matlab dataset name',
        doc='Name of the MATLAB dataset where data is stored'
    )

    connectivity = DataTypeGidAttr(
        linked_datatype=Connectivity,
        label='Large Scale Connectivity',
        doc='The Connectivity for which these measurements were made'
    )


class ConnectivityMeasureImporterForm(ABCUploaderForm):

    def __init__(self):
        super(ConnectivityMeasureImporterForm, self).__init__()

        self.data_file = TraitUploadField(ConnectivityMeasureImporterModel.data_file, '.mat', 'data_file')
        self.dataset_name = StrField(ConnectivityMeasureImporterModel.dataset_name, 'dataset_name')
        self.connectivity = TraitDataTypeSelectField(ConnectivityMeasureImporterModel.connectivity, name='connectivity')

    @staticmethod
    def get_view_model():
        return ConnectivityMeasureImporterModel

    @staticmethod
    def get_upload_information():
        return {
            'data_file': '.mat'
        }


class ConnectivityMeasureImporter(ABCUploader):
    """
    This imports a searies of conectivity measures from a .mat file
    """
    _ui_name = "ConnectivityMeasure"
    _ui_subsection = "connectivity_measure"
    _ui_description = "Import a searies of connectivity measures from a .mat file"

    def get_form_class(self):
        return ConnectivityMeasureImporterForm

    def get_output(self):
        return [ConnectivityMeasureIndex]

    @transactional
    def launch(self, view_model):
        # type: (ConnectivityMeasureImporterModel) -> [ConnectivityMeasureIndex]
        """
        Execute import operations:
        """
        try:
            data = self.read_matlab_data(view_model.data_file, view_model.dataset_name)
            measurement_count, node_count = data.shape
            connectivity = self.load_traited_by_gid(view_model.connectivity)

            if node_count != connectivity.number_of_regions:
                raise LaunchException('The measurements are for %s nodes but the selected connectivity'
                                      ' contains %s nodes' % (node_count, connectivity.number_of_regions))

            measures = []
            self.generic_attributes.user_tag_2 = "conn_%d" % node_count

            for i in range(measurement_count):
                cm_data = data[i, :]

                measure = ConnectivityMeasure()
                measure.array_data = cm_data
                measure.connectivity = connectivity
                measure.title = "Measure %d for Connectivity with %d nodes." % ((i + 1), node_count)

                cm_idx = h5.store_complete(measure, self.storage_path)
                measures.append(cm_idx)
            return measures

        except ParseException as excep:
            logger = get_logger(__name__)
            logger.exception(excep)
            raise LaunchException(excep)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from tvb.adapters.simulator.noise_forms import get_form_for_noise
from tvb.adapters.simulator.subforms_mapping import SubformsEnum, get_ui_name_to_noise_dict
from tvb.basic.neotraits.api import Attr
from tvb.core.entities.file.simulator.view_model import HeunDeterministicViewModel, HeunStochasticViewModel, \
    EulerDeterministicViewModel, EulerStochasticViewModel, RungeKutta4thOrderDeterministicViewModel, IdentityViewModel, \
    VODEViewModel, VODEStochasticViewModel, Dopri5ViewModel, Dopri5StochasticViewModel, Dop853ViewModel, \
    Dop853StochasticViewModel, IntegratorViewModel, NoiseViewModel
from tvb.core.entities.file.simulator.view_model import IntegratorStochasticViewModel
from tvb.core.neotraits.forms import Form, SelectField, FloatField


def get_integrator_to_form_dict():
    integrator_class_to_form = {
        HeunDeterministicViewModel: IntegratorForm,
        HeunStochasticViewModel: IntegratorStochasticForm,
        EulerDeterministicViewModel: IntegratorForm,
        EulerStochasticViewModel: IntegratorStochasticForm,
        RungeKutta4thOrderDeterministicViewModel: IntegratorForm,
        IdentityViewModel: IntegratorForm,
        VODEViewModel: IntegratorForm,
        VODEStochasticViewModel: IntegratorStochasticForm,
        Dopri5ViewModel: IntegratorForm,
        Dopri5StochasticViewModel: IntegratorStochasticForm,
        Dop853ViewModel: IntegratorForm,
        Dop853StochasticViewModel: IntegratorStochasticForm
    }
    return integrator_class_to_form


def get_form_for_integrator(integrator_class):
    return get_integrator_to_form_dict().get(integrator_class)


class IntegratorForm(Form):

    def get_subform_key(self):
        return SubformsEnum.INTEGRATOR.name

    def __init__(self):
        super(IntegratorForm, self).__init__()
        self.dt = FloatField(IntegratorViewModel.dt)


class IntegratorStochasticForm(IntegratorForm):
    template = 'form_fields/select_field.html'

    def __init__(self):
        super(IntegratorStochasticForm, self).__init__()
        self.noise_choices = get_ui_name_to_noise_dict()
        default_noise = list(self.noise_choices.values())[0]

        self.noise = SelectField(Attr(NoiseViewModel, label='Noise', default=default_noise), name='noise',
                                 choices=self.noise_choices, subform=get_form_for_noise(default_noise))

    def fill_trait(self, datatype):
        super(IntegratorStochasticForm, self).fill_trait(datatype)
        if self.noise.data and type(datatype.noise) != self.noise.data:
            datatype.noise = self.noise.data()

    def fill_from_trait(self, trait):
        # type: (IntegratorStochasticViewModel) -> None
        self.noise.data = trait.noise.__class__



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Here we define entities related to user and project.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Yann Gordon <yann@tvb.invalid>
"""

import datetime
import uuid

from sqlalchemy import Boolean, Integer, String, DateTime, Column, ForeignKey
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.collections import attribute_mapped_collection
from tvb.basic.logger.builder import get_logger
from tvb.basic.profile import TvbProfile
from tvb.core import utils
from tvb.core.entities.exportable import Exportable
from tvb.core.entities.transient.structure_entities import DataTypeMetaData
from tvb.core.neotraits.db import Base

LOG = get_logger(__name__)

# Constants for User Roles.
ROLE_ADMINISTRATOR = "ADMINISTRATOR"
ROLE_CLINICIAN = "CLINICIAN"
ROLE_RESEARCHER = "RESEARCHER"

USER_ROLES = [ROLE_ADMINISTRATOR, ROLE_CLINICIAN, ROLE_RESEARCHER]


class User(Base):
    """
    Contains the users informations.
    """
    __tablename__ = 'USERS'

    id = Column(Integer, primary_key=True)
    gid = Column(String, unique=True)
    username = Column(String, unique=True)
    display_name = Column(String)
    password = Column(String)
    email = Column(String, nullable=True)
    role = Column(String)
    validated = Column(Boolean)
    selected_project = Column(Integer)

    preferences = association_proxy('user_preferences', 'value',
                                    creator=lambda k, v: UserPreferences(key=k, value=v))

    def __init__(self, login, display_name, password, email=None, validated=True, role=ROLE_RESEARCHER, gid=None):
        self.username = login
        self.display_name = display_name
        self.password = password
        self.email = email
        self.validated = validated
        self.role = role
        if gid is None:
            gid = uuid.uuid4().hex
        self.gid = gid

    def __repr__(self):
        return "<USER('%s','%s','%s','%s','%s','%s', %s)>" % (self.username, self.display_name, self.password,
                                                              self.email, self.validated, self.role,
                                                              str(self.selected_project))

    def is_administrator(self):
        """Return a boolean, saying if current user has role Administrator"""
        return self.role == ROLE_ADMINISTRATOR

    def is_online_help_active(self):
        """
        This method returns True if this user should see online help.
        """
        is_help_active = True
        if UserPreferences.ONLINE_HELP_ACTIVE in self.preferences:
            flag_str = self.preferences[UserPreferences.ONLINE_HELP_ACTIVE]
            is_help_active = utils.string2bool(flag_str)

        return is_help_active

    def switch_online_help_state(self):
        """
        This method changes the state of the OnlineHelp Active flag.
        """
        self.preferences[UserPreferences.ONLINE_HELP_ACTIVE] = str(not self.is_online_help_active())

    def set_viewers_color_scheme(self, color_scheme):
        self.preferences[UserPreferences.VIEWERS_COLOR_SCHEME] = color_scheme

    def get_viewers_color_scheme(self):
        if UserPreferences.VIEWERS_COLOR_SCHEME not in self.preferences:
            self.preferences[UserPreferences.VIEWERS_COLOR_SCHEME] = "linear"

        return self.preferences[UserPreferences.VIEWERS_COLOR_SCHEME]

    def set_project_structure_grouping(self, first, second):
        k = UserPreferences.PROJECT_STRUCTURE_GROUPING
        self.preferences[k] = "%s,%s" % (first, second)

    def get_project_structure_grouping(self):
        k = UserPreferences.PROJECT_STRUCTURE_GROUPING
        if k not in self.preferences:
            self.preferences[k] = "%s,%s" % (DataTypeMetaData.KEY_STATE, DataTypeMetaData.KEY_SUBJECT)
        return self.preferences[k].split(',')


class UserPreferences(Base):
    """
    Contains the user preferences data.
    """
    __tablename__ = 'USER_PREFERENCES'

    ONLINE_HELP_ACTIVE = "online_help_active"
    VIEWERS_COLOR_SCHEME = "viewers_color_scheme"
    PROJECT_STRUCTURE_GROUPING = "project_structure_grouping"
    fk_user = Column(Integer, ForeignKey('USERS.id'), primary_key=True)
    key = Column(String, primary_key=True)
    value = Column(String)

    user = relationship(User, backref=backref("user_preferences", cascade="all, delete-orphan", lazy='joined',
                                              collection_class=attribute_mapped_collection("key")))

    def __repr__(self):
        return 'UserPreferences: %s - %s' % (self.key, self.value)


class Project(Base, Exportable):
    """
    Contains the Projects informations and who is the administrator.
    """
    __tablename__ = 'PROJECTS'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    description = Column(String)
    last_updated = Column(DateTime)
    fk_admin = Column(Integer, ForeignKey('USERS.id'))
    gid = Column(String, unique=True)
    version = Column(Integer)

    administrator = relationship(User)

    ### Transient Attributes
    # todo: remove these as they are used only to transfer info to the templates. Have incorrect values prior to the templating phase.
    # send the status to the templates differently
    operations_finished = 0
    operations_started = 0
    operations_error = 0
    operations_canceled = 0
    operations_pending = 0

    members = []

    def __init__(self, name, fk_admin, description=''):
        self.name = name
        self.fk_admin = fk_admin
        self.description = description
        self.gid = utils.generate_guid()
        self.version = TvbProfile.current.version.PROJECT_VERSION

    def refresh_update_date(self):
        """Mark entity as being changed NOW. (last_update field)"""
        self.last_updated = datetime.datetime.now()

    def __repr__(self):
        return "<Project('%s', '%s')>" % (self.name, self.fk_admin)

    def to_dict(self):
        """
        Overwrite superclass method to add required changes.
        """
        _, base_dict = super(Project, self).to_dict(excludes=['id', 'fk_admin', 'administrator', 'trait'])
        return self.__class__.__name__, base_dict

    def from_dict(self, dictionary, user_id):
        """
        Add specific attributes from a input dictionary.
        """
        self.name = dictionary['name']
        self.description = dictionary['description']
        self.last_updated = datetime.datetime.now()
        self.gid = dictionary['gid']
        self.fk_admin = user_id
        self.version = int(dictionary['version'])
        return self


class User_to_Project(Base):
    """
    Multiple Users can be members of a given Project.
    """
    __tablename__ = 'USERS_TO_PROJECTS'

    id = Column(Integer, primary_key=True)
    fk_user = Column(Integer, ForeignKey('USERS.id', ondelete="CASCADE"))
    fk_project = Column(Integer, ForeignKey('PROJECTS.id', ondelete="CASCADE"))

    def __init__(self, user, case):

        if type(user) == int:
            self.fk_user = user
        else:
            self.fk_user = user.id

        if type(case) == int:
            self.fk_project = case
        else:
            self.fk_project = case.id



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

from tvb.adapters.datatypes.db.sensors import SensorsIndex
from tvb.basic.logger.builder import get_logger
from tvb.core.adapters.exceptions import LaunchException
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import TraitUploadField, SelectField
from tvb.core.neotraits.h5 import MEMORY_STRING
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str
from tvb.datatypes.sensors import SensorsEEG, SensorsMEG, SensorsInternal


class SensorsImporterModel(UploaderViewModel):
    OPTIONS = {'EEG Sensors': SensorsEEG.sensors_type.default,
               'MEG Sensors': SensorsMEG.sensors_type.default,
               'Internal Sensors': SensorsInternal.sensors_type.default}

    sensors_file = Str(
        label='Please upload sensors file (txt or bz2 format)',
        doc='Expected a text/bz2 file containing sensor measurements.'
    )

    sensors_type = Str(
        label='Sensors type: ',
        choices=tuple(OPTIONS.values()),
        default=tuple(OPTIONS.values())[0]
    )


class SensorsImporterForm(ABCUploaderForm):

    def __init__(self):
        super(SensorsImporterForm, self).__init__()

        self.sensors_file = TraitUploadField(SensorsImporterModel.sensors_file, ('.txt', '.bz2'), 'sensors_file')
        self.sensors_type = SelectField(SensorsImporterModel.sensors_type, name='sensors_type',
                                        choices=SensorsImporterModel.OPTIONS)

    @staticmethod
    def get_view_model():
        return SensorsImporterModel

    @staticmethod
    def get_upload_information():
        return {
            'sensors_file': ('.txt', '.bz2')
        }


class SensorsImporter(ABCUploader):
    """
    Upload Sensors from a TXT file.
    """
    _ui_name = "Sensors"
    _ui_subsection = "sensors_importer"
    _ui_description = "Import Sensor locations from TXT or BZ2"

    logger = get_logger(__name__)

    def get_form_class(self):
        return SensorsImporterForm

    def get_output(self):
        return [SensorsIndex]

    def launch(self, view_model):
        # type: (SensorsImporterModel) -> [SensorsIndex]
        """
        Creates required sensors from the uploaded file.
        :returns: a list of sensors instances of the specified type

        :raises LaunchException: when
                    * no sensors_file specified
                    * sensors_type is invalid (not one of the mentioned options)
                    * sensors_type is "MEG sensors" and no orientation is specified
        """
        if view_model.sensors_file is None:
            raise LaunchException("Please select sensors file which contains data to import")

        self.logger.debug("Create sensors instance")
        if view_model.sensors_type == SensorsEEG.sensors_type.default:
            sensors_inst = SensorsEEG()
        elif view_model.sensors_type == SensorsMEG.sensors_type.default:
            sensors_inst = SensorsMEG()
        elif view_model.sensors_type == SensorsInternal.sensors_type.default:
            sensors_inst = SensorsInternal()
        else:
            exception_str = "Could not determine sensors type (selected option %s)" % view_model.sensors_type
            raise LaunchException(exception_str)

        locations = self.read_list_data(view_model.sensors_file, usecols=[1, 2, 3])

        # NOTE: TVB has the nose pointing -y and left ear pointing +x
        # If the sensors are in CTF coordinates : nose pointing +x left ear +y
        # to rotate the sensors by -90 along z uncomment below
        # locations = numpy.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]]).dot(locations.T).T
        sensors_inst.locations = locations
        sensors_inst.labels = self.read_list_data(view_model.sensors_file, dtype=MEMORY_STRING, usecols=[0])
        sensors_inst.number_of_sensors = sensors_inst.labels.size

        if isinstance(sensors_inst, SensorsMEG):
            try:
                sensors_inst.orientations = self.read_list_data(view_model.sensors_file, usecols=[4, 5, 6])
                sensors_inst.has_orientation = True
            except IndexError:
                raise LaunchException("Uploaded file does not contains sensors orientation.")

        sensors_inst.configure()
        self.logger.debug("Sensors instance ready to be stored")

        return h5.store_complete(sensors_inst, self.storage_path)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>

Data for Parameter Space Exploration view will be defined here.
The purpose of this entities is to be used in Jinja2 UI, or for populating visualizer.
"""

import json
import math
import numpy
from tvb.adapters.datatypes.db.mapped_value import DatatypeMeasureIndex
from tvb.basic.config.utils import EnhancedDictionary
from tvb.core.entities.storage import dao

KEY_GID = "Gid"
KEY_TOOLTIP = "tooltip"
LINE_SEPARATOR = "<br/>"


class PSEModel(object):
    def __init__(self, operation):
        self.operation = operation
        self.datatype_measure = self.determine_operation_result()
        self.metrics = dict()
        if self.datatype_measure:
            self.metrics = json.loads(self.datatype_measure.metrics)
        self.range1_key = None
        self.range1_value = None
        self.range2_key = None
        self.range2_value = 0
        self._extract_range_values()

    def _extract_range_values(self):
        ranges = json.loads(self.operation.range_values)
        range_keys = list(ranges.keys())
        self.range1_key = range_keys[0]
        self.range1_value = ranges[self.range1_key]

        if len(range_keys) > 1:
            self.range2_key = range_keys[1]
            self.range2_value = ranges[self.range2_key]

    def __is_float(self, value):
        is_float = True
        try:
            float(value)
        except ValueError:
            is_float = False
        return is_float

    def is_range1_float(self):
        return self.__is_float(self.range1_value)

    def is_range2_float(self):
        if not self.range2_key:
            return False
        return self.__is_float(self.range2_value)

    def _determine_type_and_label(self, value):
        if self.__is_float(value):
            return value

        label = dao.get_datatype_by_gid(value).display_name
        return label

    def get_range1_label(self):
        return self._determine_type_and_label(self.range1_value)

    def get_range2_label(self):
        if not self.range2_key:
            return ''
        return self._determine_type_and_label(self.range2_value)

    def prepare_node_info(self):
        """
        Build a dictionary with all the required information to be displayed for a given node.
        """
        KEY_NODE_TYPE = "dataType"
        KEY_OPERATION_ID = "operationId"

        node_info = dict()

        if self.operation.has_finished and self.datatype_measure is not None:
            ts = dao.get_datatype_by_gid(self.datatype_measure.fk_source_gid)
            node_info[KEY_GID] = ts.gid
            node_info[KEY_NODE_TYPE] = ts.type
            node_info[KEY_OPERATION_ID] = ts.fk_from_operation
            node_info[KEY_TOOLTIP] = self._prepare_node_text(ts)
        else:
            tooltip = "No result available. Operation is in status: %s" % self.operation.status.split('-')[1]
            node_info[KEY_TOOLTIP] = tooltip
        return node_info

    @staticmethod
    def _prepare_node_text(ts):
        """
        Prepare text to display as tooltip for a PSE circle
        :return: str
        """
        return str("Operation id: " + str(ts.fk_from_operation) + LINE_SEPARATOR +
                   "Datatype gid: " + str(ts.gid) + LINE_SEPARATOR +
                   "Datatype type: " + str(ts.type) + LINE_SEPARATOR +
                   "Datatype subject: " + str(ts.subject) + LINE_SEPARATOR +
                   "Datatype invalid: " + str(ts.invalid))

    def determine_operation_result(self):
        datatype_measure = None
        if self.operation.has_finished:
            datatypes = dao.get_results_for_operation(self.operation.id)
            if len(datatypes) > 0:
                datatype = datatypes[0]
                if datatype.type == "DatatypeMeasureIndex":
                    # Load proper entity class from DB.
                    measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid, 'gid')
                else:
                    measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid, 'fk_source_gid')
                if len(measures) > 0:
                    datatype_measure = measures[0]

        return datatype_measure


class PSEGroupModel(object):
    def __init__(self, datatype_group_gid):
        self.datatype_group = dao.get_datatype_group_by_gid(datatype_group_gid)

        if self.datatype_group is None:
            raise Exception("Selected DataTypeGroup is no longer present in the database. "
                            "It might have been remove or the specified id is not the correct one.")

        self.operation_group = dao.get_operationgroup_by_id(self.datatype_group.fk_operation_group)
        self.operations = dao.get_operations_in_group(self.operation_group.id)
        self.pse_model_list = self.parse_pse_data_for_display()
        self.all_metrics = dict()
        self._prepare_ranges_data()

    @property
    def range1_values(self):
        if self.pse_model_list[0].is_range1_float():
            return self.range1_orig_values
        return list(range(len(self.range1_orig_values)))

    @property
    def range2_values(self):
        if self.pse_model_list[0].is_range2_float():
            return self.range2_orig_values
        return list(range(len(self.range2_orig_values)))

    def parse_pse_data_for_display(self):
        pse_model_list = []
        for operation in self.operations:
            pse_model = PSEModel(operation)
            pse_model_list.append(pse_model)
        return pse_model_list

    def get_range1_key(self):
        return self.pse_model_list[0].range1_key

    def get_range2_key(self):
        return self.pse_model_list[0].range2_key

    def _prepare_ranges_data(self):
        value_to_label1 = dict()
        value_to_label2 = dict()
        for pse_model in self.pse_model_list:
            label1 = pse_model.get_range1_label()
            value_to_label1.update({pse_model.range1_value: label1})
            label2 = pse_model.get_range2_label()
            value_to_label2.update({pse_model.range2_value: label2})

        value_to_label1 = dict(sorted(value_to_label1.items()))
        value_to_label2 = dict(sorted(value_to_label2.items()))
        self.range1_orig_values = list(value_to_label1.keys())
        self.range1_labels = list(value_to_label1.values())
        self.range2_orig_values = list(value_to_label2.keys())
        self.range2_labels = list(value_to_label2.values())

    def get_all_node_info(self):
        all_node_info = dict()
        for pse_model in self.pse_model_list:
            node_info = pse_model.prepare_node_info()
            range1_val = self.range1_values[self.range1_orig_values.index(pse_model.range1_value)]
            range2_val = self.range2_values[self.range2_orig_values.index(pse_model.range2_value)]
            if not range1_val in all_node_info:
                all_node_info[range1_val] = {}
            all_node_info[range1_val][range2_val] = node_info
        return all_node_info

    def get_all_metrics(self):
        if len(self.all_metrics) == 0:
            for pse_model in self.pse_model_list:
                if pse_model.datatype_measure:
                    ts = dao.get_datatype_by_gid(pse_model.datatype_measure.fk_source_gid)
                    self.all_metrics.update({ts.gid: pse_model.metrics})
        return self.all_metrics

    def get_available_metric_keys(self):
        return list(self.pse_model_list[0].metrics)


class DiscretePSE(EnhancedDictionary):
    def __init__(self, datatype_group_gid, color_metric, size_metric, back_page):
        super(DiscretePSE, self).__init__()
        self.datatype_group_gid = datatype_group_gid
        self.min_color = float('inf')
        self.max_color = - float('inf')
        self.min_shape_size = float('inf')
        self.max_shape_size = - float('inf')
        self.has_started_ops = False
        self.status = 'started'
        self.title_x, self.title_y = "", ""
        self.values_x, self.labels_x, self.values_y, self.labels_y = [], [], [], []
        self.series_array, self.available_metrics = [], []
        self.d3_data = {}
        self.color_metric = color_metric
        self.size_metric = size_metric
        self.pse_back_page = back_page

    def prepare_individual_jsons(self):
        """
        Apply JSON.dumps on all attributes which can not be passes as they are towards UI.
        """
        self.labels_x = json.dumps(self.labels_x)
        self.labels_y = json.dumps(self.labels_y)
        self.values_x = json.dumps(self.values_x)
        self.values_y = json.dumps(self.values_y)
        self.d3_data = json.dumps(self.d3_data)
        self.has_started_ops = json.dumps(self.has_started_ops)


class PSEDiscreteGroupModel(PSEGroupModel):

    def __init__(self, datatype_group_gid, color_metric, size_metric, back_page):
        super(PSEDiscreteGroupModel, self).__init__(datatype_group_gid)
        self.color_metric = color_metric
        self.size_metric = size_metric
        self.determine_default_metrics()
        self.pse_context = DiscretePSE(datatype_group_gid, self.color_metric, self.size_metric, back_page)
        self.fill_pse_context()
        self.prepare_display_data()

    def determine_default_metrics(self):
        if self.color_metric is None and self.size_metric is None:
            metrics = self.get_available_metric_keys()
            if len(metrics) > 0:
                self.color_metric = metrics[0]
                self.size_metric = metrics[0]
            if len(metrics) > 1:
                self.size_metric = metrics[1]

    def fill_pse_context(self):
        sizes_array = [metrics[self.size_metric] for metrics in self.get_all_metrics().values()]
        colors_array = [metrics[self.color_metric] for metrics in self.get_all_metrics().values()]
        if len(sizes_array) > 0:
            self.pse_context.min_shape_size = numpy.min(sizes_array)
            self.pse_context.max_shape_size = numpy.max(sizes_array)
            self.pse_context.min_color = numpy.min(colors_array)
            self.pse_context.max_color = numpy.max(colors_array)
        self.pse_context.available_metrics = self.get_available_metric_keys()
        self.pse_context.title_x = self.get_range1_key()
        self.pse_context.title_y = self.get_range2_key()
        self.pse_context.values_x = self.range1_values
        self.pse_context.values_y = self.range2_values
        self.pse_context.labels_x = self.range1_labels
        self.pse_context.labels_y = self.range2_labels

    def _prepare_coords_json(self, val1, val2):
        return '{"x":' + str(val1) + ', "y":' + str(val2) + '}'

    def prepare_display_data(self):
        d3_data = self.get_all_node_info()
        series_array = []

        for val1 in d3_data.keys():
            for val2 in d3_data[val1].keys():
                current = d3_data[val1][val2]
                coords = self._prepare_coords_json(val1, val2)
                datatype_gid = None
                if KEY_GID in current:
                    # This means the operation was finished
                    datatype_gid = current[KEY_GID]

                color_weight, shape_type_1 = self.__get_color_weight(datatype_gid)
                if (shape_type_1 is not None) and (datatype_gid is not None):
                    d3_data[val1][val2][KEY_TOOLTIP] += LINE_SEPARATOR + " Color metric has NaN values"
                current['color_weight'] = color_weight

                radius, shape_type_2 = self.__get_node_size(datatype_gid, len(self.range1_values),
                                                            len(self.range2_values))
                if (shape_type_2 is not None) and (datatype_gid is not None):
                    d3_data[val1][val2][KEY_TOOLTIP] += LINE_SEPARATOR + " Size metric has NaN values"
                symbol = shape_type_1 or shape_type_2

                series_array.append(self.__get_node_json(symbol, radius, coords))

        self.pse_context.d3_data = d3_data
        self.pse_context.series_array = self.__build_series_json(series_array)

    @staticmethod
    def __build_series_json(list_of_series):
        """ Given a list with all the data points, build the final FLOT JSON. """
        final_json = "["
        for i, value in enumerate(list_of_series):
            if i:
                final_json += ","
            final_json += value
        final_json += "]"
        return final_json

    @staticmethod
    def __get_node_json(symbol, radius, coords):
        """
        For each data point entry, build the FLOT specific JSON.
        """
        series = '{"points": {'
        if symbol is not None:
            series += '"symbol": "' + symbol + '", '
        series += '"radius": ' + str(radius) + '}, '
        series += '"coords": ' + coords + '}'
        return series

    def __get_node_size(self, datatype_gid, range1_length, range2_length):
        """
        Computes the size of the shape used for representing the dataType with GID given.
        """
        min_size, max_size = self.__get_boundaries(range1_length, range2_length)
        if datatype_gid is None:
            return max_size / 2.0, "cross"
        node_info = self.get_all_metrics()[datatype_gid]

        if self.size_metric in node_info:
            valid_metric = True
            try:
                if math.isnan(float(node_info[self.size_metric])) or math.isinf(float(node_info[self.size_metric])):
                    valid_metric = False
            except ValueError:
                valid_metric = False

            if valid_metric:
                shape_weight = node_info[self.size_metric]
                values_range = self.pse_context.max_shape_size - self.pse_context.min_shape_size
                shape_range = max_size - min_size
                if values_range != 0:
                    return min_size + (shape_weight - self.pse_context.min_shape_size) / float(
                        values_range) * shape_range, None
                else:
                    return min_size, None
            else:
                return max_size / 2.0, "cross"
        return max_size / 2.0, None

    def __get_color_weight(self, datatype_gid):
        """
        Returns the color weight of the shape used for representing the dataType which id equal to 'datatype_gid'.
        :param: datatype_gid - It should exists into the 'datatype_indexes' dictionary.
        """
        if datatype_gid is None:
            return 0, "cross"
        node_info = self.get_all_metrics()[datatype_gid]
        valid_metric = True

        if self.color_metric in node_info:
            try:
                if math.isnan(float(node_info[self.color_metric])) or math.isinf(float(node_info[self.color_metric])):
                    valid_metric = False
            except ValueError:
                valid_metric = False
            if valid_metric:
                return node_info[self.color_metric], None
            else:
                return 0, "cross"
        return 0, None

    @staticmethod
    def __get_boundaries(range1_length, range2_length):
        """
        Returns the MIN and the max values of the interval from
        which may be set the size of a certain shape.
        """
        # the arrays 'intervals' and 'values' should have the same size
        intervals = [0, 3, 5, 8, 10, 20, 30, 40, 50, 60, 90, 110, 120]
        values = [(10, 50), (10, 40), (10, 33), (8, 25), (5, 15), (4, 10),
                  (3, 8), (2, 6), (2, 5), (1, 4), (1, 3), (1, 2), (1, 2)]

        max_length = max([range1_length, range2_length])
        if max_length <= intervals[0]:
            return values[0]
        elif max_length >= intervals[-1]:
            return values[-1]
        else:
            for i, interval in enumerate(intervals):
                if max_length <= interval:
                    return values[i - 1]



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from collections import OrderedDict
from tvb.datatypes.surfaces import CORTICAL, INNER_SKULL, OUTER_SKIN, OUTER_SKULL, EEG_CAP, FACE, WHITE_MATTER, Surface
from sqlalchemy import Column, Integer, ForeignKey, String, Float, Boolean
from tvb.core.entities.model.model_datatype import DataType

ALL_SURFACES_SELECTION = OrderedDict()
ALL_SURFACES_SELECTION['Cortical Surface'] = CORTICAL
ALL_SURFACES_SELECTION['Brain Skull'] = INNER_SKULL
ALL_SURFACES_SELECTION['Skull Skin'] = OUTER_SKULL
ALL_SURFACES_SELECTION['Skin Air'] = OUTER_SKIN
ALL_SURFACES_SELECTION['EEG Cap'] = EEG_CAP
ALL_SURFACES_SELECTION['Face Surface'] = FACE
ALL_SURFACES_SELECTION['White Matter Surface'] = WHITE_MATTER


class SurfaceIndex(DataType):
    id = Column(Integer, ForeignKey(DataType.id), primary_key=True)

    surface_type = Column(String, nullable=False)
    valid_for_simulations = Column(Boolean, nullable=False)
    number_of_vertices = Column(Integer, nullable=False)
    number_of_triangles = Column(Integer, nullable=False)
    number_of_edges = Column(Integer, nullable=False)
    bi_hemispheric = Column(Boolean, nullable=False)
    edge_mean_length = Column(Float, nullable=False)
    edge_min_length = Column(Float, nullable=False)
    edge_max_length = Column(Float, nullable=False)

    def fill_from_has_traits(self, datatype):
        # type: (Surface)  -> None
        super(SurfaceIndex, self).fill_from_has_traits(datatype)
        self.surface_type = datatype.surface_type
        self.valid_for_simulations = datatype.valid_for_simulations
        self.number_of_vertices = datatype.number_of_vertices
        self.number_of_triangles = datatype.number_of_triangles
        self.number_of_edges = datatype.number_of_edges
        self.bi_hemispheric = datatype.bi_hemispheric
        self.edge_mean_length = datatype.edge_mean_length
        self.edge_min_length = datatype.edge_min_length
        self.edge_max_length = datatype.edge_max_length

    def get_subtype_attr(self):
        return self.surface_type

    @property
    def display_name(self):
        """
        Overwrite from superclass and subtype
        """
        previous = "Surface"
        return previous + " - " + str(self.surface_type)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

from tvb.adapters.datatypes.db.time_series import *
from tvb.adapters.datatypes.h5.surface_h5 import SurfaceH5
from tvb.adapters.datatypes.h5.time_series_h5 import TimeSeriesH5, TimeSeriesSurfaceH5
from tvb.adapters.visualizers.eeg_monitor import EegMonitor
from tvb.adapters.visualizers.sensors import prepare_mapped_sensors_as_measure_points_params
from tvb.adapters.visualizers.sensors import prepare_sensors_as_measure_points_params, function_sensors_to_surface
from tvb.adapters.visualizers.surface_view import ensure_shell_surface, SurfaceURLGenerator, ABCSurfaceDisplayer
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.adapters.abcdisplayer import URLGenerator
from tvb.core.entities.storage import dao
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import TraitDataTypeSelectField
from tvb.core.neotraits.view_model import DataTypeGidAttr, ViewModel
from tvb.datatypes.surfaces import CORTICAL, EEG_CAP, Surface

MAX_MEASURE_POINTS_LENGTH = 600


class BrainViewerModel(ViewModel):
    time_series = DataTypeGidAttr(
        linked_datatype=TimeSeries,
        label='Time Series (Region or Surface)'
    )

    shell_surface = DataTypeGidAttr(
        linked_datatype=Surface,
        required=False,
        label='Shell Surface',
        doc='Surface to be displayed semi-transparently as overlay, for visual navigation purposes only.'
    )


class BrainViewerForm(ABCAdapterForm):

    def __init__(self):
        super(BrainViewerForm, self).__init__()
        self.time_series = TraitDataTypeSelectField(BrainViewerModel.time_series, name='time_series',
                                                    conditions=self.get_filters())
        self.shell_surface = TraitDataTypeSelectField(BrainViewerModel.shell_surface, name='shell_surface')

    @staticmethod
    def get_view_model():
        return BrainViewerModel

    @staticmethod
    def get_required_datatype():
        return TimeSeriesIndex

    @staticmethod
    def get_input_name():
        return 'time_series'

    @staticmethod
    def get_filters():
        return FilterChain(
            fields=[FilterChain.datatype + '.time_series_type', FilterChain.datatype + '.has_surface_mapping'],
            operations=["in", "=="], values=[['TimeSeriesRegion', 'TimeSeriesSurface'], True])


class BrainViewer(ABCSurfaceDisplayer):
    """
    Interface between the 3D view of the Brain Cortical Surface and TVB framework.
    This viewer will build the required parameter dictionary that will be sent to the HTML / JS for further processing,
    having as end result a brain surface plus activity that will be displayed in 3D.
    """
    _ui_name = "Brain Activity Visualizer"
    PAGE_SIZE = 500

    def get_form_class(self):
        return BrainViewerForm

    def get_required_memory_size(self, view_model):
        # type: (BrainViewerModel) -> numpy.ndarray
        """
        Assume one page doesn't get 'dumped' in time and it is highly probably that
        two consecutive pages will be in the same time in memory.
        """
        time_series = self.load_entity_by_gid(view_model.time_series)
        overall_shape = time_series.get_data_shape()
        used_shape = (overall_shape[0] / (self.PAGE_SIZE * 2.0), overall_shape[1], overall_shape[2], overall_shape[3])
        return numpy.prod(used_shape) * 8.0

    def generate_preview(self, view_model, figure_size=None):
        """
        Generate the preview for the burst page
        """
        time_series_index = self.load_entity_by_gid(view_model.time_series)
        self.populate_surface_fields(time_series_index)

        url_vertices, url_normals, url_lines, url_triangles, url_region_map = \
            SurfaceURLGenerator.get_urls_for_rendering(self.surface_h5, self.region_map_gid)

        params = self.retrieve_measure_points_params(time_series_index)
        base_adapter_url, time_urls = self._prepare_data_slices(time_series_index)

        with h5.h5_file_for_index(time_series_index) as time_series_h5:
            assert isinstance(time_series_h5, TimeSeriesH5)
            min_val, max_val = time_series_h5.get_min_max_values()

        if self.surface_gid and self.region_map_gid:
            boundary_url = SurfaceURLGenerator.get_url_for_region_boundaries(self.surface_gid, self.region_map_gid,
                                                                             self.stored_adapter.id)
        else:
            boundary_url = ''

        params.update(urlVertices=json.dumps(url_vertices), urlTriangles=json.dumps(url_triangles),
                      urlLines=json.dumps(url_lines), urlNormals=json.dumps(url_normals),
                      urlRegionMap=json.dumps(url_region_map), urlRegionBoundaries=boundary_url,
                      base_adapter_url=base_adapter_url, isOneToOneMapping=self.one_to_one_map,
                      minActivity=min_val, maxActivity=max_val)

        normalization_factor = figure_size[0] / 800
        if figure_size[1] / 600 < normalization_factor:
            normalization_factor = figure_size[1] / 600
        params['width'] = figure_size[0] * normalization_factor
        params['height'] = figure_size[1] * normalization_factor

        if self.surface_h5:
            self.surface_h5.close()

        return self.build_display_result("brain/portlet_preview", params)

    def launch(self, view_model):
        # type: (BrainViewerModel) -> dict
        """
        Build visualizer's page.
        """
        time_series_index = self.load_entity_by_gid(view_model.time_series)
        shell_surface_index = None
        if view_model.shell_surface:
            shell_surface_index = self.load_entity_by_gid(view_model.shell_surface)
        params = self.compute_parameters(time_series_index, shell_surface_index)
        return self.build_display_result("brain/view", params, pages=dict(controlPage="brain/controls"))

    def populate_surface_fields(self, time_series_index):
        """
        To be overwritten for populating fields: one_to_one_map/connectivity/region_map/surface fields
        """

        self.one_to_one_map = isinstance(time_series_index, TimeSeriesSurfaceIndex)

        if self.one_to_one_map:
            self.PAGE_SIZE /= 10
            surface_gid = time_series_index.fk_surface_gid
            surface_index = dao.get_datatype_by_gid(surface_gid)
            region_map_indexes = dao.get_generic_entity(RegionMappingIndex, surface_gid, 'fk_surface_gid')
            if len(region_map_indexes) < 1:
                region_map_index = None
                connectivity_index = None
            else:
                region_map_index = region_map_indexes[0]
                connectivity_index = dao.get_datatype_by_gid(region_map_index.fk_connectivity_gid)
        else:
            connectivity_index = dao.get_datatype_by_gid(time_series_index.fk_connectivity_gid)

            if time_series_index.fk_region_mapping_gid:
                region_map_index = dao.get_datatype_by_gid(time_series_index.fk_region_mapping_gid)
            else:
                region_map_indexes = dao.get_generic_entity(RegionMappingIndex, connectivity_index.gid,
                                                            'fk_connectivity_gid')
                region_map_index = region_map_indexes[0]

            surface_index = dao.get_datatype_by_gid(region_map_index.fk_surface_gid)

        self.connectivity_index = connectivity_index
        self.region_map_gid = None if region_map_index is None else region_map_index.gid
        self.surface_gid = None if surface_index is None else surface_index.gid
        self.surface_h5 = None if surface_index is None else h5.h5_file_for_index(surface_index)

    def retrieve_measure_points_params(self, time_series):
        """
        To be overwritten method, for retrieving the measurement points (region centers, EEG sensors).
        """
        if self.connectivity_index is None:
            self.measure_points_no = 0
            return {'urlMeasurePoints': [],
                    'urlMeasurePointsLabels': [],
                    'noOfMeasurePoints': 0}

        connectivity_gid = self.connectivity_index.gid
        measure_points = SurfaceURLGenerator.build_h5_url(connectivity_gid, 'get_centres')
        measure_points_labels = SurfaceURLGenerator.build_h5_url(connectivity_gid, 'get_region_labels')
        self.measure_points_no = self.connectivity_index.number_of_regions

        return {'urlMeasurePoints': measure_points,
                'urlMeasurePointsLabels': measure_points_labels,
                'noOfMeasurePoints': self.measure_points_no}

    def compute_parameters(self, time_series, shell_surface=None):
        """
        Create the required parameter dictionary for the HTML/JS viewer.

        :rtype: `dict`
        :raises Exception: when
                    * number of measure points exceeds the maximum allowed
                    * a Face object cannot be found in database

        """
        self.populate_surface_fields(time_series)

        url_vertices, url_normals, url_lines, url_triangles, url_region_map = SurfaceURLGenerator.get_urls_for_rendering(
            self.surface_h5, self.region_map_gid)
        hemisphere_chunk_mask = self.surface_h5.get_slices_to_hemisphere_mask()

        params = self.retrieve_measure_points_params(time_series)

        if not self.one_to_one_map and self.measure_points_no > MAX_MEASURE_POINTS_LENGTH:
            raise Exception("Max number of measure points " + str(MAX_MEASURE_POINTS_LENGTH) + " exceeded.")

        time_series_h5 = h5.h5_file_for_index(time_series)
        assert isinstance(time_series_h5, TimeSeriesH5)
        base_adapter_url, time_urls = self._prepare_data_slices(time_series)
        min_val, max_val = time_series_h5.get_min_max_values()
        legend_labels = self._compute_legend_labels(min_val, max_val)

        state_variables = time_series.get_labels_for_dimension(1)

        if self.surface_gid and self.region_map_gid:
            boundary_url = SurfaceURLGenerator.get_url_for_region_boundaries(self.surface_gid, self.region_map_gid,
                                                                             self.stored_adapter.id)
        else:
            boundary_url = ''

        shell_surface = ensure_shell_surface(self.current_project_id, shell_surface)
        params.update(dict(title="Cerebral Activity: " + time_series.title, isOneToOneMapping=self.one_to_one_map,
                           urlVertices=json.dumps(url_vertices), urlTriangles=json.dumps(url_triangles),
                           urlLines=json.dumps(url_lines), urlNormals=json.dumps(url_normals),
                           urlRegionMap=json.dumps(url_region_map), base_adapter_url=base_adapter_url,
                           time=json.dumps(time_urls), minActivity=min_val, maxActivity=max_val,
                           legendLabels=legend_labels, labelsStateVar=state_variables,
                           labelsModes=list(range(time_series.data_length_4d)), extended_view=False,
                           shellObject=self.prepare_shell_surface_params(shell_surface, SurfaceURLGenerator),
                           biHemispheric=self.surface_h5.bi_hemispheric.load(),
                           hemisphereChunkMask=json.dumps(hemisphere_chunk_mask),
                           pageSize=self.PAGE_SIZE, urlRegionBoundaries=boundary_url,
                           measurePointsLabels=self.get_space_labels(time_series_h5),
                           measurePointsTitle=time_series.title))

        params.update(self.build_params_for_subselectable_ts(time_series_h5))

        time_series_h5.close()
        if self.surface_h5:
            self.surface_h5.close()

        return params

    @staticmethod
    def _prepare_mappings(mappings_dict):
        """
        Get full mapping dictionary between the original vertices and multiple slices (for WebGL compatibility).
        """
        prepared_mappings = []
        for key in mappings_dict:
            this_mappings = []
            vert_map_dict = mappings_dict[key]
            vertices_indexes = vert_map_dict['indices']
            this_mappings.append(vertices_indexes[0].tolist())
            for i in range(1, len(vertices_indexes)):
                if vertices_indexes[i][0] == vertices_indexes[i][1]:
                    this_mappings.append(vertices_indexes[i][0])
                else:
                    for index in range(vertices_indexes[i][0], vertices_indexes[i][1] + 1):
                        this_mappings.append(index)
            prepared_mappings.append(this_mappings)
        return prepared_mappings

    @staticmethod
    def _compute_legend_labels(min_val, max_val, nr_labels=5, min_nr_dec=3):
        """
        Compute rounded labels for MIN and MAX values such that decimals will show a difference between them.
        """
        if len(str(min_val).split('.')) == 2:
            min_integer, min_decimals = str(min_val).split('.')
        else:
            min_integer, min_decimals = [str(int(min_val)), ""]
        if len(str(max_val).split('.')) == 2:
            max_integer, max_decimals = str(max_val).split('.')
        else:
            max_integer, max_decimals = [str(int(max_val)), ""]
        idx = min_nr_dec
        if len(min_decimals) < min_nr_dec or len(max_decimals) < min_nr_dec:
            processed_min_val = float(min_val)
            processed_max_val = float(max_val)
        elif min_integer != max_integer:
            processed_min_val = float(min_integer + '.' + min_decimals[:min_nr_dec])
            processed_max_val = float(max_integer + '.' + max_decimals[:min_nr_dec])
        else:
            for idx, val in enumerate(min_decimals):
                if idx < len(max_decimals) or val != max_decimals[idx]:
                    break
            processed_min_val = float(min_integer + '.' + min_decimals[:idx])
            processed_max_val = float(max_integer + '.' + max_decimals[:idx])
        value_diff = (processed_max_val - processed_min_val) / (nr_labels + 1)
        inter_values = [round(processed_min_val + value_diff * i, idx) for i in range(nr_labels, 0, -1)]
        return [processed_max_val] + inter_values + [processed_min_val]

    def _prepare_data_slices(self, time_series_index):
        """
        Prepare data URL for retrieval with slices of timeSeries activity and Time-Line.
        :returns: [activity_urls], [timeline_urls]
                 Currently timeline_urls has just one value, as on client is loaded entirely anyway.
        """
        time_series_gid = time_series_index.gid
        activity_base_url = URLGenerator.build_url(self.stored_adapter.id, 'read_data_page_split', time_series_gid, "")
        time_urls = [SurfaceURLGenerator.build_h5_url(time_series_gid, 'read_time_page',
                                                      parameter="current_page=0;page_size=" +
                                                                str(time_series_index.data_length_1d))]
        return activity_base_url, time_urls

    def read_data_page_split(self, time_series_gid, from_idx, to_idx, step=None, specific_slices=None):
        with h5.h5_file_for_gid(time_series_gid) as time_series_h5:
            assert isinstance(time_series_h5, TimeSeriesH5)
            basic_result = time_series_h5.read_data_page(from_idx, to_idx, step, specific_slices)

            if not isinstance(time_series_h5, TimeSeriesSurfaceH5):
                return basic_result.tolist()
            surface_gid = time_series_h5.surface.load()

        result = []
        with h5.h5_file_for_gid(surface_gid) as surface_h5:
            assert isinstance(surface_h5, SurfaceH5)
            number_of_split_slices = surface_h5.number_of_split_slices.load()
            if number_of_split_slices <= 1:
                result.append(basic_result.tolist())
            else:
                for slice_number in range(surface_h5.number_of_split_slices):
                    start_idx, end_idx = surface_h5.get_slice_vertex_boundaries(slice_number)
                    result.append(basic_result[:, start_idx:end_idx].tolist())

        return result


class DualBrainViewerModel(ViewModel):
    time_series = DataTypeGidAttr(
        linked_datatype=TimeSeries,
        label='Time Series'
    )

    projection_surface = DataTypeGidAttr(
        linked_datatype=Surface,
        required=False,
        label='Projection Surface',
        doc='A surface on which to project the results. When missing, the first EEGCap is taken. '
            'This parameter is ignored when InternalSensors measures.'
    )

    shell_surface = DataTypeGidAttr(
        linked_datatype=Surface,
        required=False,
        label='Shell Surface',
        doc='Wrapping surface over the internal sensors, to be displayed '
            'semi-transparently, for visual purposes only.'
    )


class DualBrainViewerForm(ABCAdapterForm):

    def __init__(self):
        super(DualBrainViewerForm, self).__init__()
        self.time_series = TraitDataTypeSelectField(DualBrainViewerModel.time_series,
                                                    name='time_series',
                                                    conditions=self.get_filters())
        self.projection_surface = TraitDataTypeSelectField(DualBrainViewerModel.projection_surface,
                                                           name='projection_surface')
        self.shell_surface = TraitDataTypeSelectField(DualBrainViewerModel.shell_surface,
                                                      name='shell_surface')

    @staticmethod
    def get_view_model():
        return DualBrainViewerModel

    @staticmethod
    def get_required_datatype():
        return TimeSeriesIndex

    @staticmethod
    def get_input_name():
        return 'time_series'

    @staticmethod
    def get_filters():
        return FilterChain(
            fields=[FilterChain.datatype + '.time_series_type', FilterChain.datatype + '.has_surface_mapping'],
            operations=["in", "=="],
            values=[['TimeSeriesEEG', 'TimeSeriesSEEG', 'TimeSeriesMEG', 'TimeSeriesRegion'], True])


class DualBrainViewer(BrainViewer):
    """
    Visualizer merging Brain 3D display and EEG lines display.
    """
    _ui_name = "Brain Activity Viewer in 3D and 2D"
    _ui_subsection = "brain_dual"

    def get_form_class(self):
        return DualBrainViewerForm

    def populate_surface_fields(self, time_series_index):
        """
        Prepares the urls from which the client may read the data needed for drawing the surface.
        """

        if isinstance(time_series_index, TimeSeriesRegionIndex):
            BrainViewer.populate_surface_fields(self, time_series_index)
            return

        self.one_to_one_map = False
        self.region_map_gid = None
        self.connectivity_index = None

        if self.surface_index is None:
            eeg_cap = dao.get_generic_entity(SurfaceIndex, EEG_CAP, "surface_type")
            if len(eeg_cap) < 1:
                raise Exception("No EEG Cap Surface found for display!")
            self.surface_index = eeg_cap[0]

        self.surface_gid = self.surface_index.gid
        self.surface_h5 = h5.h5_file_for_index(self.surface_index)

    def retrieve_measure_points_params(self, time_series):

        if isinstance(time_series, TimeSeriesRegionIndex):
            return BrainViewer.retrieve_measure_points_params(self, time_series)

        sensors_index = dao.get_datatype_by_gid(time_series.fk_sensors_gid)
        self.measure_points_no = sensors_index.number_of_sensors

        if isinstance(time_series, TimeSeriesEEGIndex):
            return prepare_mapped_sensors_as_measure_points_params(sensors_index, self.surface_index,
                                                                   self.stored_adapter.id)

        return prepare_sensors_as_measure_points_params(sensors_index)

    def sensors_to_surface(self, sensors_gid, surface_to_map_gid):
        # Method needs to be defined on the adapter, to be called from JS
        return function_sensors_to_surface(sensors_gid, surface_to_map_gid)

    def launch(self, view_model):
        # type: (DualBrainViewerModel) -> dict

        time_series_index = self.load_entity_by_gid(view_model.time_series)
        self.surface_index = None
        shell_surface_index = None

        if view_model.projection_surface:
            self.surface_index = self.load_entity_by_gid(view_model.projection_surface)

        if view_model.shell_surface:
            shell_surface_index = self.load_entity_by_gid(view_model.shell_surface)

        if isinstance(time_series_index, TimeSeriesSEEGIndex):
            shell_surface_index = ensure_shell_surface(self.current_project_id, shell_surface_index, CORTICAL)

        params = BrainViewer.compute_parameters(self, time_series_index, shell_surface_index)
        eeg_monitor = EegMonitor()
        eeg_monitor.storage_path = self.storage_path
        params.update(eeg_monitor.compute_parameters(time_series_index, is_extended_view=True))

        params['isOneToOneMapping'] = False
        params['brainViewerTemplate'] = 'view.html'

        if isinstance(time_series_index, TimeSeriesSEEGIndex):
            params['brainViewerTemplate'] = "internal_view.html"
            # Mark as None since we only display shell face and no point to load these as well
            params['urlVertices'] = None
            params['isSEEG'] = True

        return self.build_display_result("brain/extendedview", params,
                                         pages=dict(controlPage="brain/extendedcontrols",
                                                    channelsPage="commons/channel_selector.html"))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import threading
import urllib.parse as parse
from http.server import BaseHTTPRequestHandler, HTTPServer
from time import sleep

from tvb.interfaces.rest.client.client_decorators import handle_response
from tvb.interfaces.rest.client.main_api import MainApi
from tvb.interfaces.rest.commons.dtos import ProjectDto
from tvb.interfaces.rest.commons.status_codes import HTTP_STATUS_REDIRECT, HTTP_STATUS_BAD_REQUEST
from tvb.interfaces.rest.commons.strings import RestLink, FormKeyInput, RestNamespace, Strings


class UserApi(MainApi):
    @handle_response
    def login(self, code, redirect_uri):
        response = self.secured_request().post(self.build_request_url(RestLink.LOGIN.compute_url(True)), json={
            FormKeyInput.CODE.value: code,
            FormKeyInput.REDIRECT_URI.value: redirect_uri
        })
        return response

    @handle_response
    def logout(self):
        response = self.secured_request().delete(self.build_request_url(RestLink.LOGIN.compute_url(True)), json={
            FormKeyInput.KEYCLOAK_REFRESH_TOKEN.value: self.refresh_token,
        })
        return response

    @handle_response
    def get_projects_list(self):
        response = self.secured_request().get(self.build_request_url(RestLink.PROJECTS.compute_url(True)))
        return response, ProjectDto

    @handle_response
    def get_users(self, page):
        return self.secured_request().get(self.build_request_url(RestNamespace.USERS.value), params={
            Strings.PAGE_NUMBER: page
        })

    @handle_response
    def create_project(self, project_name, project_description):
        return self.secured_request().post(self.build_request_url(RestLink.PROJECTS.compute_url(True)), json={
            FormKeyInput.CREATE_PROJECT_NAME.value: project_name,
            FormKeyInput.CREATE_PROJECT_DESCRIPTION.value: project_description
        })

    @handle_response
    def _get_urls(self, redirect_uri):
        return self.secured_request().get(self.build_request_url(RestLink.USEFUL_URLS.compute_url(True)), params={
            FormKeyInput.REDIRECT_URI.value: redirect_uri
        })

    def browser_login(self, login_callback_port, open_browser_function):
        host = "127.0.0.1"
        redirect_uri = "http://{}:{}".format(host, login_callback_port)

        # Fetch account and authorization urls
        urls = self._get_urls(redirect_uri)
        account_url = urls[Strings.ACCOUNT_URL.value]
        auth_url = urls[Strings.AUTH_URL.value]

        class LoginCallbackApp(BaseHTTPRequestHandler):
            """
            An handler of request for the keycloak redirect
            """

            authorization_code = None
            has_error = False

            def do_GET(self):
                try:
                    LoginCallbackApp.authorization_code = parse.parse_qs(self.path[2:])['code'][0]
                    self.send_response(HTTP_STATUS_REDIRECT)
                    self.send_header('Location', account_url)
                    self.end_headers()
                except KeyError:
                    LoginCallbackApp.has_error = True
                    self.send_response(HTTP_STATUS_BAD_REQUEST)

        # Configure and start a basic HTTP server
        httpd = HTTPServer((host, login_callback_port), LoginCallbackApp)
        th = threading.Thread(target=httpd.serve_forever)
        th.start()
        open_browser_function(auth_url)
        while LoginCallbackApp.authorization_code is None and not LoginCallbackApp.has_error:
            # Wait until login is performed
            sleep(1)
        httpd.shutdown()

        # Ask keycloak server for tokens using the authorization code
        return self.login(code=LoginCallbackApp.authorization_code, redirect_uri=redirect_uri)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import json

from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.orm import relationship
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.adapters.datatypes.h5.mapped_value_h5 import ValueWrapper
from tvb.core.entities.file.simulator.datatype_measure_h5 import DatatypeMeasure
from tvb.core.entities.model.model_datatype import DataType


class ValueWrapperIndex(DataType):
    """
    Class to wrap a singular value storage in DB.
    """
    id = Column(Integer, ForeignKey(DataType.id), primary_key=True)
    data_value = Column(String)
    data_type = Column(String, default='unknown')
    data_name = Column(String)

    @property
    def display_name(self):
        """ Simple String to be used for display in UI."""
        return "Value Wrapper - " + self.data_name + " : " + str(self.data_value) + " (" + str(self.data_type) + ")"

    def fill_from_has_traits(self, datatype):
        # type: (ValueWrapper)  -> None
        super(ValueWrapperIndex, self).fill_from_has_traits(datatype)
        self.data_value = datatype.data_value
        self.data_type = datatype.data_type
        self.data_name = datatype.data_name


class DatatypeMeasureIndex(DataType):
    """
    Class to hold the metric for a previous stored DataType.
    E.g. Measure (single value) for any TimeSeries resulted in a group of Simulations
    """
    id = Column(Integer, ForeignKey(DataType.id), primary_key=True)
    # Actual measure (dictionary Algorithm: single Value) serialized
    metrics = Column(String)
    # DataType for which the measure was computed.
    fk_source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid), nullable=False)
    source = relationship(TimeSeriesIndex, foreign_keys=fk_source_gid, primaryjoin=TimeSeriesIndex.gid == fk_source_gid)

    @property
    def display_name(self):

        result = super(DatatypeMeasureIndex, self).display_name

        if self.metrics is not None:
            metrics = json.loads(self.metrics)
            for entry, metric_value in metrics.items():
                result = result + " -- " + entry + ' : ' + str(metric_value)

        return result

    def fill_from_has_traits(self, datatype):
        # type: (DatatypeMeasure)  -> None
        super(DatatypeMeasureIndex, self).fill_from_has_traits(datatype)
        self.metrics = json.dumps(datatype.metrics)
        self.fk_source_gid = datatype.analyzed_datatype.gid.hex



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
"""

import json
import uuid
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.core.adapters.exceptions import LaunchException, ParseException
from tvb.adapters.uploaders.gifti.parser import GIFTIParser
from tvb.basic.logger.builder import get_logger
from tvb.adapters.datatypes.h5.time_series_h5 import TimeSeriesSurfaceH5
from tvb.adapters.datatypes.db.time_series import TimeSeriesSurfaceIndex
from tvb.core.neotraits.forms import TraitUploadField, TraitDataTypeSelectField
from tvb.core.neotraits.db import prepare_array_shape_meta
from tvb.core.neocom import h5
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str, DataTypeGidAttr
from tvb.datatypes.surfaces import Surface


class GIFTITimeSeriesImporterModel(UploaderViewModel):
    data_file = Str(
        label='Please select file to import (.gii)'
    )

    surface = DataTypeGidAttr(
        linked_datatype=Surface,
        label='Brain Surface',
        doc='The Brain Surface used to generate imported TimeSeries.'
    )


class GIFTITimeSeriesImporterForm(ABCUploaderForm):

    def __init__(self):
        super(GIFTITimeSeriesImporterForm, self).__init__()

        self.data_file = TraitUploadField(GIFTITimeSeriesImporterModel.data_file, '.gii', 'data_file')
        surface_conditions = FilterChain(fields=[FilterChain.datatype + '.surface_type'], operations=["=="],
                                         values=['Cortical Surface'])
        self.surface = TraitDataTypeSelectField(GIFTITimeSeriesImporterModel.surface, name='surface',
                                                conditions=surface_conditions)

    @staticmethod
    def get_view_model():
        return GIFTITimeSeriesImporterModel

    @staticmethod
    def get_upload_information():
        return {
            'data_file': '.gii'
        }


class GIFTITimeSeriesImporter(ABCUploader):
    """
        This importer is responsible for import of a TimeSeries from GIFTI format (XML file)
        and store them in TVB.
    """
    _ui_name = "TimeSeries GIFTI"
    _ui_subsection = "gifti_timeseries_importer"
    _ui_description = "Import TimeSeries from GIFTI"

    def get_form_class(self):
        return GIFTITimeSeriesImporterForm

    def get_output(self):
        return [TimeSeriesSurfaceIndex]

    def launch(self, view_model):
        # type: (GIFTITimeSeriesImporterModel) -> [TimeSeriesSurfaceIndex]
        """
        Execute import operations:
        """
        if view_model.surface is None:
            raise LaunchException("No surface selected. Please initiate upload again and select a brain surface.")

        parser = GIFTIParser(self.storage_path, self.operation_id)
        try:
            partial_time_series, gifti_data_arrays = parser.parse(view_model.data_file)

            ts_idx = TimeSeriesSurfaceIndex()
            ts_h5_path = h5.path_for(self.storage_path, TimeSeriesSurfaceH5, ts_idx.gid)

            ts_h5 = TimeSeriesSurfaceH5(ts_h5_path)
            # todo : make sure that write_time_slice is not required here
            for data_array in gifti_data_arrays:
                ts_h5.write_data_slice([data_array.data])

            ts_h5.store(partial_time_series, scalars_only=True, store_references=False)
            ts_h5.gid.store(uuid.UUID(ts_idx.gid))

            ts_data_shape = ts_h5.read_data_shape()
            surface = self.load_entity_by_gid(view_model.surface)
            if surface.number_of_vertices != ts_data_shape[1]:
                msg = "Imported time series doesn't have values for all surface vertices. Surface has %d vertices " \
                      "while time series has %d values." % (surface.number_of_vertices, ts_data_shape[1])
                raise LaunchException(msg)
            else:
                ts_h5.surface.store(uuid.UUID(surface.gid))
                ts_idx.fk_surface_gid = surface.gid
            ts_h5.close()

            ts_idx.sample_period_unit = partial_time_series.sample_period_unit
            ts_idx.sample_period = partial_time_series.sample_period
            ts_idx.sample_rate = partial_time_series.sample_rate
            ts_idx.labels_ordering = json.dumps(partial_time_series.labels_ordering)
            ts_idx.labels_dimensions = json.dumps(partial_time_series.labels_dimensions)
            ts_idx.data_ndim = len(ts_data_shape)
            ts_idx.data_length_1d, ts_idx.data_length_2d, ts_idx.data_length_3d, ts_idx.data_length_4d = prepare_array_shape_meta(
                ts_data_shape)

            return [ts_idx]

        except ParseException as excep:
            logger = get_logger(__name__)
            logger.exception(excep)
            raise LaunchException(excep)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.adapters.uploaders.zip_surface.parser import ZipSurfaceParser
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import Attr
from tvb.core.adapters.exceptions import LaunchException
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.adapters.datatypes.db.surface import SurfaceIndex, ALL_SURFACES_SELECTION
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import TraitUploadField, SelectField, BoolField
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str
from tvb.datatypes.surfaces import make_surface, center_vertices


class ZIPSurfaceImporterModel(UploaderViewModel):
    uploaded = Str(
        label='Surface file (zip)'
    )

    surface_type = Str(
        choices=tuple(ALL_SURFACES_SELECTION.values()),
        default=tuple(ALL_SURFACES_SELECTION.values())[0],
        label='Surface type'
    )

    zero_based_triangles = Attr(
        field_type=bool,
        required=False,
        default=True,
        label='Zero based triangles'
    )

    should_center = Attr(
        field_type=bool,
        required=False,
        label='Center surface using vertex means along axes'
    )


class ZIPSurfaceImporterForm(ABCUploaderForm):

    def __init__(self):
        super(ZIPSurfaceImporterForm, self).__init__()
        self.uploaded = TraitUploadField(ZIPSurfaceImporterModel.uploaded, '.zip', 'uploaded')
        self.surface_type = SelectField(ZIPSurfaceImporterModel.surface_type, 'surface_type',
                                        choices=ALL_SURFACES_SELECTION)
        self.zero_based_triangles = BoolField(ZIPSurfaceImporterModel.zero_based_triangles, name='zero_based_triangles')
        self.should_center = BoolField(ZIPSurfaceImporterModel.should_center, name='should_center')

    @staticmethod
    def get_view_model():
        return ZIPSurfaceImporterModel

    @staticmethod
    def get_upload_information():
        return {
            'uploaded': '.zip'
        }


class ZIPSurfaceImporter(ABCUploader):
    """
    Handler for uploading a Surface Data archive, with files holding
    vertices, normals and triangles to represent a surface data.
    """

    _ui_name = "Surface ZIP"
    _ui_subsection = "zip_surface_importer"
    _ui_description = "Import a Surface from ZIP"
    logger = get_logger(__name__)

    def get_form_class(self):
        return ZIPSurfaceImporterForm

    def get_output(self):
        return [SurfaceIndex]

    @staticmethod
    def _make_surface(surface_type):

        result = make_surface(surface_type)

        if result is not None:
            return result

        exception_str = "Could not determine surface type (selected option %s)" % surface_type
        raise LaunchException(exception_str)

    def launch(self, view_model):
        # type: (ZIPSurfaceImporterModel) -> [SurfaceIndex]
        """
        Execute import operations: unpack ZIP and build Surface object as result
        :raises LaunchException: when
                * `uploaded` is missing
                * `surface_type` is invalid
        :raises RuntimeError: when triangles contain an invalid vertex index
        """
        if view_model.uploaded is None:
            raise LaunchException("Please select ZIP file which contains data to import")

        self.logger.debug(
            "Start to import surface: '%s' from file: %s" % (view_model.surface_type, view_model.uploaded))
        try:
            zip_surface = ZipSurfaceParser(view_model.uploaded)
        except IOError:
            exception_str = "Did not find the specified ZIP at %s" % view_model.uploaded
            raise LaunchException(exception_str)

        # Detect and instantiate correct surface type
        self.logger.debug("Create surface instance")
        surface = self._make_surface(view_model.surface_type)
        surface.zero_based_triangles = view_model.zero_based_triangles
        if view_model.should_center:
            vertices = center_vertices(zip_surface.vertices)
        else:
            vertices = zip_surface.vertices
        surface.vertices = vertices
        if len(zip_surface.normals) != 0:
            surface.vertex_normals = zip_surface.normals
        if view_model.zero_based_triangles:
            surface.triangles = zip_surface.triangles
        else:
            surface.triangles = zip_surface.triangles - 1

        if zip_surface.bi_hemispheric:
            self.logger.info("Hemispheres detected")

        surface.hemisphere_mask = zip_surface.hemisphere_mask
        surface.compute_triangle_normals()

        # Now check if the triangles of the surface are valid
        triangles_min_vertex = numpy.amin(surface.triangles)
        if triangles_min_vertex < 0:
            if triangles_min_vertex == -1 and not view_model.zero_based_triangles:
                raise LaunchException("Triangles contain a negative vertex index. Maybe you have a ZERO based surface.")
            else:
                raise LaunchException("Your triangles contain a negative vertex index: %d" % triangles_min_vertex)

        no_of_vertices = len(surface.vertices)
        triangles_max_vertex = numpy.amax(surface.triangles)
        if triangles_max_vertex >= no_of_vertices:
            if triangles_max_vertex == no_of_vertices and view_model.zero_based_triangles:
                raise LaunchException("Your triangles contain an invalid vertex index: %d. "
                                      "Maybe your surface is NOT ZERO Based." % triangles_max_vertex)
            else:
                raise LaunchException("Your triangles contain an invalid vertex index: %d." % triangles_max_vertex)

        validation_result = surface.validate()

        if validation_result.warnings:
            self.add_operation_additional_info(validation_result.summary())

        surface.configure()
        self.logger.debug("Surface ready to be stored")

        return h5.store_complete(surface, self.storage_path)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>    
"""

import json
import six
import tvb.core.entities.model.model_operation as model
from tvb.core.entities.filters.chain import FilterChain


class StaticFiltersFactory(object):
    """
    Factory class to build lists with static used filters through the application.
    """
    RELEVANT_VIEW = "Relevant view"
    FULL_VIEW = "Full view"

    @staticmethod
    def build_datatype_filters(selected=RELEVANT_VIEW, single_filter=None):
        """
        Return all visibility filters for data structure page, or only one filter.
        """
        filters = {StaticFiltersFactory.FULL_VIEW: FilterChain(StaticFiltersFactory.FULL_VIEW),
                   StaticFiltersFactory.RELEVANT_VIEW: FilterChain(StaticFiltersFactory.RELEVANT_VIEW,
                                                                   [FilterChain.datatype + '.visible'],
                                                                   [True], operations=["=="])}
        if selected is None or len(selected) == 0:
            selected = StaticFiltersFactory.RELEVANT_VIEW
        if selected in filters:
            filters[selected].selected = True
        if single_filter is not None:
            if single_filter in filters:
                return filters[single_filter]
            else:
                # We have some custom filter to build
                return StaticFiltersFactory._build_custom_filter(single_filter)
        return list(filters.values())

    @staticmethod
    def _build_custom_filter(filter_data):
        """
        Param filter_data should be at this point a dictionary of the form:
        {'type' : 'filter_type', 'value' : 'filter_value'}
        If 'filter_type' is not handled just return None.
        """
        filter_data = json.loads(filter_data)
        if filter_data['type'] == 'from_burst':
            return FilterChain('Burst', ['BurstConfiguration.id'],
                               [filter_data['value']], operations=["=="])
        return None

    @staticmethod
    def build_operations_filters(simulation_algorithm, logged_user_id):
        """
        :returns: list of filters that can be applied on Project View Operations page.
        """
        new_filters = []

        # Filter by algorithm / categories
        new_filter = FilterChain("Omit Views", [FilterChain.algorithm_category + '.display'],
                                 [False], operations=["=="])
        new_filters.append(new_filter)

        new_filter = FilterChain("Only Upload", [FilterChain.algorithm_category + '.rawinput'],
                                 [True], operations=["=="])
        new_filters.append(new_filter)
        if simulation_algorithm is not None:
            new_filter = FilterChain("Only Simulations", [FilterChain.algorithm + '.id'],
                                     [simulation_algorithm.id], operations=["=="])
            new_filters.append(new_filter)

        # Filter by operation status
        filtered_statuses = {model.STATUS_STARTED: "Only Running",
                             model.STATUS_ERROR: "Only with Errors",
                             model.STATUS_CANCELED: "Only Canceled",
                             model.STATUS_FINISHED: "Only Finished",
                             model.STATUS_PENDING: "Only Pending"}
        for status, title in six.iteritems(filtered_statuses):
            new_filter = FilterChain(title, [FilterChain.operation + '.status'], [status], operations=["=="])
            new_filters.append(new_filter)

        # Filter by author
        new_filter = FilterChain("Only mine", [FilterChain.operation + '.fk_launched_by'],
                                 [logged_user_id], operations=["=="])
        new_filters.append(new_filter)

        # Filter by other flags
        new_filter = FilterChain("Only relevant", [FilterChain.operation + '.visible'], [True], operations=["=="])
        new_filter.selected = True
        new_filters.append(new_filter)

        return new_filters



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

from tvb.basic.profile import TvbProfile
from tvb.basic.config import stored
from tvb.core.code_versions.base_classes import UpdateManager
import tvb.core.code_versions.code_update_scripts as code_versions



class CodeUpdateManager(UpdateManager):
    """
    A manager that goes through all the scripts that are newer than the version number 
    written in the .tvb.basic.config.setting configuration file.
    """


    def __init__(self):
        super(CodeUpdateManager, self).__init__(code_versions, TvbProfile.current.version.CODE_CHECKED_TO_VERSION,
                                                TvbProfile.current.version.REVISION_NUMBER)


    def run_update_script(self, script_name):
        """
        Add specific code after every update script.
        """
        super(CodeUpdateManager, self).run_update_script(script_name)
        # After each update mark the update in cfg file. 
        # In case one update script fails, the ones before will not be repeated.
        TvbProfile.current.manager.add_entries_to_config_file(
            {stored.KEY_LAST_CHECKED_CODE_VERSION: script_name.split('_')[0]})


    def run_all_updates(self):
        """
        Upgrade the code to current version. 
        Go through all update scripts with lower SVN version than the current running version.
        """
        if TvbProfile.is_first_run():
            ## We've just started with a clean TVB. No need to upgrade anything.
            return

        super(CodeUpdateManager, self).run_all_updates()

        if self.checked_version < self.current_version:
            TvbProfile.current.manager.add_entries_to_config_file(
                {stored.KEY_LAST_CHECKED_CODE_VERSION: TvbProfile.current.version.REVISION_NUMBER})
        
        
        
        